/*-
 * Copyright (c) 2024 Varnish Software AS
 * All rights reserved.
 *
 * Author: Stephane Cance <stephane.cance@varnish-software.com>
 *
 * Implementation of the core rate broker which provides token
 * buckets working against shared rates between nodes.
 *
 * Each account holds a token bucket and is part of a collection,
 * it is assumed that account belonging to the same collection share
 * an ordered communication stream and behave in a similar way. All
 * publish callbacks are called from the same single thread without
 * holding any locks.
 */

#ifdef RTBROKER_H_INCLUDED
# error "loopy includes"
#endif /* RTBROKER_H_INCLUDED */
#define RTBROKER_H_INCLUDED

/*
 * UIDs identify nodes within the communication group.
 */
#define RTBROKER_UID_SIZE		8

typedef union rtbroker_uid {
	uint64_t        u64[RTBROKER_UID_SIZE / sizeof(uint64_t)];
	uint8_t         u8[RTBROKER_UID_SIZE / sizeof(uint8_t)];
} rtbroker_uid;

/*
 * COLIDs identify collections.
 */
#define RTBROKER_COLID_SIZE		32

typedef union rtbroker_colid {
	uint64_t        u64[RTBROKER_COLID_SIZE / sizeof(uint64_t)];
	uint8_t         u8[RTBROKER_COLID_SIZE / sizeof(uint8_t)];
} rtbroker_colid;

/*
 * ACCIDs identify accounts within a collection.
 */
#define RTBROKER_ACCID_SIZE		32

typedef union rtbroker_accid {
	uint64_t        u64[RTBROKER_ACCID_SIZE / sizeof(uint64_t)];
	uint8_t         u8[RTBROKER_ACCID_SIZE / sizeof(uint8_t)];
} rtbroker_accid;

/*
 * Configuration
 */

struct rtbroker_ctx_config {
	struct VSC_lck			*vsc_mtx;
	struct VSC_cond			*vsc_cond;
	vtim_dur			max_event_jitter;

#define RTBROKER_CTX_CNT(nm)	uint64_t *nm##p;
#include "rtbroker/tbl_rtbroker_stats.h"
};

typedef void rtbroker_publish_heartbeat_f(const rtbroker_colid *colid,
    const rtbroker_uid *uid, uint64_t seqid, vtim_dur expiry_ttl,
    uint32_t accounts_with_leases, void *userdata);
typedef void rtbroker_publish_gone_f(const rtbroker_colid *colid,
    const rtbroker_uid *uid, uint64_t seqid, void *userdata);
typedef void rtbroker_account_done_f(const rtbroker_colid *colid,
    const rtbroker_accid *accid, void *userdata);

struct rtbroker_collection_config {
	struct VSC_lck			*vsc_mtx;

	vtim_dur			heartbeat_ttl;
	vtim_dur			heartbeat_grace;

	uint32_t			discovery_assumed_participants;

	rtbroker_publish_heartbeat_f	*publish_heartbeat;
	rtbroker_publish_gone_f		*publish_gone;
	rtbroker_account_done_f		*account_done;

#define RTBROKER_COL_CNT(nm)	uint64_t *nm##p;
#include "rtbroker/tbl_rtbroker_stats.h"
};

typedef void rtbroker_publish_lease_f(const rtbroker_colid *colid,
    const rtbroker_uid *uid, uint64_t seqid, const rtbroker_accid *accid,
    double max_rate, double rate, vtim_dur expiry_ttl, void *userdata);

struct rtbroker_account_config {
	struct VSC_lck			*vsc_mtx;

	vtim_dur			lease_ttl;
	vtim_dur			lease_grace;
	vtim_dur			stability_ttl;

	vtim_dur			tb_credit_duration;
	vtim_dur			tb_constrained_grace;

	rtbroker_publish_lease_f	*publish_lease;

#define RTBROKER_ACC_CNT(nm)	uint64_t *nm##p;
#define RTBROKER_ACC_GAUGE(nm)	uint64_t *nm##p;
#include "rtbroker/tbl_rtbroker_stats.h"
};

struct rtbroker_ctx;
struct rtbroker_collection;
struct rtbroker_account;

/*
 * The rate limiter maintains the state of collections and accounts by running
 * a background thread. The uid must outlive the context.
 * Returns NULL if the background thread cannot be created.
 */
struct rtbroker_ctx *rtbroker_ctx_new(const rtbroker_uid *uid,
    struct rtbroker_ctx_config *cfg);
void rtbroker_ctx_free(struct rtbroker_ctx *ctx);

/*
 * wait for the background thread to be idle, this does not mean
 * that the event queue is empty just that it currently cannot deliver
 * mode events. This is used for testing purposes and note that only
 * a single concurrent call to rtbroker_ctx_wait_idle() is supported.
 */
void rtbroker_ctx_wait_idle(struct rtbroker_ctx *ctx);

/* stop the background thread, proper shutdown procedure is:
 *  - stop delivery of messages
 *  - stop background thread (e.g. call rtbroker_ctx_stop())
 *  - free all accounts and all collections
 *  - shutdown ability to publish messages
 *  - free context
 */
void rtbroker_ctx_stop(struct rtbroker_ctx *ctx);

/* start the background thread after it was stopped */
int rtbroker_ctx_start(struct rtbroker_ctx *ctx);

/*
 * create collections
 * ctx and colid must outlive the collection.
 * collection is initialized with a multiplier of 1.0.
 */
struct rtbroker_collection *rtbroker_collection_new(struct rtbroker_ctx *ctx,
    const rtbroker_colid *colid, const struct rtbroker_collection_config *cfg,
    void *userdata);
/*
 * can only be called when collection has no account left or
 * if the context has been stopped.
 */
void rtbroker_collection_free(struct rtbroker_collection *col);

/*
 * notifies collection that the communication status to peers has changed for
 * all of its accounts. The provided multiplier is multiplied to any aquired
 * local rate (it is not communicated to peers at all). This may have a
 * significant cost as all accounts will react quickly to the new multiplier and
 * drop the stable state. Also this triggers a discovery restart.
 * It is appropriate to call this when it is known that communication either has
 * broken down or is back up.
 */
void rtbroker_collection_communication_change(struct rtbroker_collection *col,
    double multiplier);

/*
 * create accounts
 * col and accid must outlive the account.
 */
struct rtbroker_account *rtbroker_account_new(struct rtbroker_collection *col,
    const rtbroker_accid *accid, double maximum_rate,
    const struct rtbroker_account_config *cfg, void *userdata);
/*
 * can only be called when the account is done or if the context
 * has been stopped
 */
void rtbroker_account_free(struct rtbroker_account *account);

/*
 * Accounts should be kept if it is currently tracking remote state, this
 * allows to free an account after receiving an account_done notification.
 * If account is freed this returns non-zero, otherwise this returns zero.
 */
int rtbroker_account_free_if_done(struct rtbroker_account *account);

/* communication receive handlers */
void rtbroker_collection_receive_heartbeat(struct rtbroker_collection *col,
    const rtbroker_uid *from_uid, uint64_t seqid, vtim_dur expiry_ttl,
    uint32_t accounts_with_leases);
void rtbroker_collection_receive_gone(struct rtbroker_collection *col,
    const rtbroker_uid *from_uid, uint64_t seqid);
void rtbroker_account_receive_lease(struct rtbroker_account *account,
    const rtbroker_uid *from_uid, uint64_t seqid, double max_rate, double rate,
    vtim_dur expiry_ttl);

void rtbroker_account_set_max_rate(struct rtbroker_account *account,
    double max_rate);
double rtbroker_account_get_shared_max_rate(struct rtbroker_account *acc);
void rtbroker_account_set_credit_duration(struct rtbroker_account *acc,
    vtim_dur tb_credit_duration);

/* spending from an account, returns 0 when spend operation did not succeed */
int rtbroker_account_spend(struct rtbroker_account *account, double amount,
    int force);
