/*-
 * Copyright (c) 2025 Varnish Software AS
 * All rights reserved.
 *
 * Author: Stephane Cance <stephane.cance@varnish-software.com>
 */
#ifdef VADIS_LIB_H_INCLUDED
#  error "vadis.h invluded multiple times"
#endif
#define VADIS_LIB_H_INCLUDED

#include <hiredis/hiredis.h>

struct vadis_mux;
struct vadis_channel;

struct vadis_channel_counters
{
	uint64_t    *c_n_conn_attemptp;
	uint64_t    *c_n_readp;
	uint64_t    *c_n_writep;
	uint64_t    *c_n_pingp;
};

enum vadis_channel_event
{
	VADIS_CHAN_ERR_CONN,
	VADIS_CHAN_ERR_COMPAT,
	VADIS_CHAN_ERR_AUTH,
	VADIS_CHAN_ERR_SELECT,
	VADIS_CHAN_ERR_TIMEOUT,
	VADIS_CHAN_ERR_IO,
	VADIS_CHAN_CONNECTED,
	VADIS_CHAN_READY,
};

struct vadis_mux *vadis_mux_create(const char *name, struct VSC_lck *vsc_mtx);
void vadis_mux_destroy(struct vadis_mux **muxp);

/* indicate event loop to exit */
void vadis_mux_end_event_loop(struct vadis_mux *mux);

/*
 * Before calling vadis_mux_event_loop() allows blocking of some signals
 * to avoid disturbing redis communication.
 */
void vadis_block_signals(sigset_t *oldset);

/*
 * enter the event loop, this does not return until another thread
 * invokes vadis_mux_end_event_loop().
 */
void vadis_mux_event_loop(struct vadis_mux *mux);

/*
 * For all error types, str is non-NULL and describes the error. An error
 * event implies that the connection will be severed and all appended command
 * if any won't get any replies (must be retried). Connection retries are
 * automatic and the first retry is immediate followed by an exponential
 * back-off up to max_reconnect_timeout.
 * For the CONNECTED event, str is the detected redis server version if
 * available. Note that this event occurs before authentication.
 * For the READY event, str is NULL.
 */
typedef void vadis_channel_event_f(struct vadis_channel *chan,
    enum vadis_channel_event evt, const char *str, void *userdata);
typedef int vadis_channel_reply_f(struct vadis_channel *chan,
    const redisReply *reply, void *userdata);

/*
 * Create a channel, note that the channel is immediately starting to
 * attempt connecting to the redis server.
 * note: it is not legal to destroy a channel from the event loop thread
 */
struct vadis_channel *vadis_channel_create(struct vadis_mux *mux,
    const char *name, const char *def, int db, const char *user,
    const char *pass, vtim_dur conn_timeout, vtim_dur cmd_timeout,
    vtim_dur max_reconnect_timeout, vadis_channel_event_f *event_cb,
    vadis_channel_reply_f *reply_cb, void *userdata,
    struct vadis_channel_counters *counters);
void vadis_channel_destroy(struct vadis_channel **chanp);

/* slightly more convenient reply inspection API */
const char *vadis_reply_type_str(const redisReply *reply, int *has_strp);
int vadis_reply_is_type(const redisReply *reply, int type);
int vadis_reply_is_ok(const redisReply *reply);
int vadis_reply_is_queued(const redisReply *reply);
int vadis_reply_is_wrongtype_err(const redisReply *r);
const redisReply *vadis_reply_map_field(const redisReply *r, const char *name);

/*
 * Warning: only a subset of printf(3) format is available here and all
 * separators between command fields must be present in the format string.
 * The choice of the attribute is only there to avoid some common mistakes,
 * but it is not enough to ensure safety.
 * This is only legal from the event_cb in the same thread as it is invoked
 * and only for a READY event.
 */
void v_printflike_(2, 3) vadis_channel_append_cmd(struct vadis_channel *chan,
    const char *format, ...);

/*
 * Send any command previously appended, no further appends can be done until
 * the event_cb is invoked again.
 * This is only legal from the event_cb in the same thread as it is invoked
 * and only for a READY event.
 * If there are no queued commands, the channel will be notified if still
 * ready after idle_tmout is passed. If idle_tmout is negative no notification
 * happends until channel changes state or a nudge happens.
 */
void vadis_channel_pipeline_complete(struct vadis_channel *chan,
    vtim_dur idle_tmout);

/*
 * If channel is in ready state but last pipeline was empty, wake it up by
 * invoking the event callback with READY event again.
 * Can be invoked from any thread.
 */
void vadis_channel_nudge_if_ready(struct vadis_channel *chan);
