/*-
 * Copyright (c) 2025 Varnish Software AS
 * All rights reserved.
 *
 * Author: Stephane Cance <stephane.cance@varnish-software.com>
 */
#ifdef VADISKV_LIB_H_INCLUDED
#  error "vadiskv.h invluded multiple times"
#endif
#define VADISKV_LIB_H_INCLUDED

struct vadis_kv_reply_tracker
{
	unsigned			magic;
#define VADIS_KV_REPLY_TRACKER_MAGIC	    0x321d12f4
	uint32_t			recid;
	uint32_t			cmdid;
	uint32_t			remain;
	char				**buffers;
	uint32_t			n_buffers;
	uint32_t			capa_buffers;
	uint32_t			next_act;
	uint32_t			n_rec_retry;
};

/*
 * Utility function which can be used to append all commands that
 * kvstore journal records requires.
 * This invokes vadis_channel_append_cmd() one or more times per record, the
 * same restrictions apply to this function.
 * A limitation of this call is that keys, fields and values in the record
 * must be nul-terminated strings and that the corresponding length indicates
 * the size of the string (e.g. it includes the terminating zero byte).
 * If is_retry is non-zero, only append records marked in the actions
 * with the KVSTORE_JRNL_WILL_RETRY bit.
 * This returns the number of commands appended and initializes *trk with a
 * valid reply tracker.
 */
uint32_t vadis_channel_append_kv_records(struct vadis_channel *chan,
    const struct kvstore_jrnl_rec *recs, size_t n_recs, int is_retry,
    struct vadis_kv_reply_tracker *trk);

/*
 * Initializer for a tracker only suitable for handling failures when the
 * commands were not appended.
 */
void vadis_kv_reply_failure_tracker_init(struct vadis_kv_reply_tracker *trk);

/*
 * cleanup tracker and frees all memory allocated in it
 * this is only safe to invoke once all records are acted upon regardless
 * of retries (a record to be retried never references memory held by the
 * tracker).
 */
void vadis_kv_reply_tracker_fini(struct vadis_kv_reply_tracker *trk);

/*
 * allocate a block of uninitialized memory of sz bytes and keep track of it
 * in the tracker which will free it upon calling vadis_kv_reply_tracker_fini.
 */
void *vadis_kv_reply_tracker_alloc(struct vadis_kv_reply_tracker *trk,
    size_t sz);

/* Returns non-zero if all replies were handled. */
int vadis_kv_reply_tracker_is_done(const struct vadis_kv_reply_tracker *trk);
/* Returns the number of records marked for a retry */
size_t vadis_kv_reply_tracker_retries(const struct vadis_kv_reply_tracker *trk);

/*
 * Handle a reply for commands appended in vadis_channel_append_kv_records().
 * Typically invoked from a vadis_channel_reply_f callback and this returns a
 * non-zero value when a reset should be done (as the callback should).
 * Records are acted upon in the kvstore automatically when appropriate.
 * Note that any previously set actions in the records are cleared during
 * handling which should be left entirely to this call.
 * Uses vadis_kv_reply_tracker_alloc() when memory is needed.
 * If can_retry is non-zero any error is translated as a success regardless
 * of errors (giving up on consistency leading to divergence).
 * If reply is NULL, then the current record and all following records part
 * of the current tracker will be considered failed, use this if no further
 * replies can be received. A non-zero value will be returned.
 */
int vadis_kv_reply_tracker_handle_reply(struct vadis_kv_reply_tracker *trk,
    struct kvstore *kv, struct kvstore_jrnl_rec *recs, size_t n_recs,
    int is_retry, int can_retry, const redisReply *reply);
